//----------------------------------------------------------------------
// pretranslate.c
//
// this program reads t9n.txt, and writes a "pretranslated" version of it
// by applying search and replace with the translations in pretrans.txt 
// (the latter a file generated by the frequentstrings tool and subsequently
// edited).
//
// The output from this program can then be used as a basis for the 
// actual help file translation, instead of t9n.txt, which could much speed 
// up translation of OpenOffice.org/Star Office help files into another 
// language since many words are already translated. 
//
// (c) 2002 S. Brouwer (simonbr@openoffice.org)
// You are welcome to use, distribute, modify etc. this software
// provided that this copyright message is preserved in it and you 
// give due credit to its author. 
//----------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <process.h>
#include <stdlib.h>
#include <ctype.h>

unsigned char iswordchartab[]={		// all isgraphic() characters, except
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	// the characters !,.:;?
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
0,0,1,1,1,1,1,1,1,1,1,1,0,1,0,1,
1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
};

#define iswordchar(c) (iswordchartab[c])


#define MAXNUMBEROFSTRINGS 69871	// this is the actual number of translatable strings in t9n.txt
#define BUFSIZE 0x10000	// 


typedef struct 
{
	char *string;			// pointer to "translatable" string from t9n.txt
	unsigned int word_count;// number of words in it
	unsigned int line;		// the line the string is in in t9n.txt
} instring_t;



void getstrings (void);
int hash (const char *word);
void calcnumberofwords (void);
void *myalloc (unsigned int size);
void search_replace (void);
void mergeresult(void);
void rehash (unsigned int nwords);
void dotrans (char *orig_string, char *start, char *tobereplaced, char *replacement);


FILE *infile, *infile2, *outfile;

int line, pos; 

char buf[BUFSIZE];
char buf2[BUFSIZE];
char buf3[BUFSIZE];

instring_t allstrings[MAXNUMBEROFSTRINGS];	// all translatable strings

unsigned int numentries;	// stat variable to keep track of unique words/sequences

unsigned int heapused;	// stat variable to keep track of memory use

unsigned int maxwordsinstring;	// maximum number of words in any string

unsigned int numberofstrings;


void
main (int argc, char *argv[])
{
	numberofstrings=0;

	heapused=0;
	numentries=0;

	if ((infile = fopen("t9n.txt", "rt"))== NULL)
	{
		printf("Error: Cannot open input file t9n.txt\n");
		exit (1);
	}

	if ((infile2 = fopen("pretrans.txt", "rt"))== NULL)
	{
		printf("Error: Cannot open input file pretrans.txt\n");
		exit (1);
	}

	if ((outfile = fopen("t9n_p.txt", "wt"))== NULL)
	{
		printf("Error: Cannot open output file t9n_p.txt\n");
		exit (1);
	}
	
	puts ("reading input file");

	getstrings();

	puts ("calculating number of words in input strings");
	calcnumberofwords();
	
	search_replace();
	
	mergeresult();

    fclose(infile);
	fclose(outfile);
	exit (0);
}


void
getstrings (void)
{
	int line=0;
	int x=0;
	
	numberofstrings=0;

	for (;;)
	{
		fgets(buf, BUFSIZE-2, infile);
		if (feof (infile))
			return;
		line++;

		if (strncmp("****", buf, 4)==0)
		{
			fgets(buf, BUFSIZE-2, infile);
			if (feof (infile))
				return;
			line++;
			
			numberofstrings++;
			if (numberofstrings>MAXNUMBEROFSTRINGS)
			{
				puts ("ERROR: more strings in input file than expected");
				exit (1);
			}

			allstrings[x].string=myalloc (strlen (buf)+1);
			heapused+=strlen (buf)+1+16;
			allstrings[x].line=line;
			strcpy (allstrings[x++].string, buf);
		}
	}
}

void
search_replace (void)
{
	unsigned int numwords=0;
	int skip;
	unsigned int i;
	unsigned char *cp, *cp2, *cp3, *cp4;
	unsigned char *lookp, *findp;
	unsigned int progress=0;

	for (;;)
	{
		if (progress++>=100)
		{
			printf(".");
			progress=0;
		}


		fgets(buf, BUFSIZE-2, infile2);
		if (feof (infile2))
			return;

		if (strncmp("####", buf, 4)==0)
		{
			if (sscanf (buf, "#### strings with %i words", &numwords)!=1)
			{
				puts ("ERROR: could not read number of words");
				exit (1);
			}
			printf ("\npre-translating substrings with %i words", numwords);
			continue;
		}
		if (numwords==0)
			continue;		

		skip=0;	
		cp=buf;
		while (*cp>=' ')
		{
			if (*cp++=='>')			// find if there is a translation
			{
				cp3=cp-1;

				while (!iswordchar(*cp))
				{
					cp++;
					if (*cp<' ')	// no translation: skip
					{
						skip=1;
						break;
					}
				}
				if (!skip)
				{
					*cp3--='\0';			// terminate "original" string
					if (!iswordchar(*cp3))   // accept space or no space in front of '>'
						*cp3--='\0';
					cp3=buf;
					while (*cp3++!=']')
						if (*cp3<' ')
						{
							printf ("ERROR: wrong format in %s\n", buf);
							exit (1);
						}
					if (*cp3++!=' ')
					{
						printf ("ERROR: wrong format in %s\n", buf);
						exit (1);
					}
					// cp3 now points to the "original" string
					// cp points to the translation
					cp2=cp;
					while (*cp2>=' ')
						cp2++;
					*cp2='\0';			// remove newline at end

					// make modified translation string in buf2
					buf2[0]='>';
					cp2=buf2+1;
					while (*cp)
					{
						if (iswordchar (*cp))
							*cp2++=*cp++;
						else
						{
							*cp2++=*cp++;
							*cp2++='>';		// mark translated string as such
						}
					}
					*cp2='\0';
					
					// cp and cp2 can be used, cp3 points to the "original" string

					for (i=0; i<numberofstrings; i++) 
					{
						if (allstrings[i].word_count>=numwords)
						{
							buf3[0]='\0';	// assemble translated string in buf3
							
							// cp: copy pointer in original string
							// cp2: copy pointer in translated string
							// findp: points to original string found in translatable string
							// lookp: points to from where original string must be looked for
							lookp=findp=cp=allstrings[i].string;
							cp2=buf3;
	
							for(;;)
							{
								if ((findp=strstr(lookp, cp3))==NULL) // orig string not found (anymore)
									break;
								
								if (iswordchar(*(findp-1))&&(findp!=allstrings[i].string))
								{
									if (*(findp+1)=='\0')
										break;
									lookp=findp+1;
									continue;	// only replace on word boundaries
								}
					
								if (iswordchar(*(findp+strlen(cp3))))
								{
									if (*(findp+1)=='\0')
										break;
									lookp=findp+1;
									continue;	// only replace on word boundaries
								}

								while (cp!=findp)	// copy up to found orig string
									*cp2++=*cp++;
								cp4=buf2;
								while (*cp4)		// add the translated string
									*cp2++=*cp4++;
								cp=lookp=findp+strlen(cp3);
							}
							if (buf3[0]!=0)
							{
								while (*cp)
									*cp2++=*cp++;
								*cp2='\0';
								heapused-=(strlen(allstrings[i].string)+1+16);
								free (allstrings[i].string);
								allstrings[i].string=myalloc(strlen(buf3)+1);
								strcpy(allstrings[i].string, buf3);
							}
						}
					}
				}
			}
		}
	}
}


// this function re-reads the input file t9n.txt, 
// and replaces all the second lines of the line pairs by the translated 
// strings, filtering out the '>' markers. 
void mergeresult(void)
{
	unsigned int line=0;
	int i=0;
	unsigned char c, *cp;

	puts ("\nwriting results to file...");

	rewind (infile);

	for (;;)
	{
		fgets(buf, BUFSIZE-2, infile);
		if (feof (infile))
			return;
		line++;

		fputs (buf, outfile); 
		if (strncmp("****", buf, 4)==0)
		{
			fgets(buf, BUFSIZE-2, infile);
			if (feof (infile))
				return;
			line++;
			
			if (allstrings[i].line!=line)
			{
				printf ("ERROR: lines mixed up at line %i", line); 
				exit (1);
			}		
			
			cp=allstrings[i].string;
			
			while (*cp)
			{
				if ((c=*cp++)!='>')
					fputc (c, outfile); 
			}
			fputc ('\n', outfile);
			
			i++;
		}
	}
}


// this function calculates the number of words in all the strings from input file
void
calcnumberofwords (void)
{
	unsigned char *cp;
	unsigned int i;
	
	maxwordsinstring=0;

	for (i=0;i<numberofstrings;i++)
	{
		allstrings[i].word_count=0;
		cp=allstrings[i].string;
		for (;;)
		{	
			if (*cp=='\0')
				break;
			if (!iswordchar(*cp))
			{
				cp++;
				continue;
			}
			while (iswordchar(*cp))
			{
				cp++;
			}
			allstrings[i].word_count++;
		}
		if (maxwordsinstring<allstrings[i].word_count)
			maxwordsinstring=allstrings[i].word_count;
	}
	printf ("maximum number of words in any string= %i\n", maxwordsinstring);
}


// a wrapper for malloc with error check and tracking of heap size used
void *myalloc (unsigned int size)
{
	void *result;

	if ((result=malloc(size))==NULL)
	{
		puts ("ERROR: malloc returned NULL");
		exit (1);
	}
	heapused+=size+16; // estimated overhead of malloc: 16 bytes
	return (result);
}

